/* $Id: fetch.c,v 1.18 2007-03-23 11:42:33 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <assert.h>
#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <curl/curl.h>

#include "nlpcrawl.h"

/*
 * This is the actual "crawl" utility that  downloads content from the
 * Internet.  In order to custom-filter headers and/or body, consult the
 * curl_header and curl_data functions.
 */

/**
 * Structured carried around by every extant connection.  It should be
 * assumed that all fields are valid during the lifetime of a data
 * transfer.
 */
struct ecurl {
	CURL		*curl;
	/**< The active connection object.
	 */
	struct nlpcdbr	rec;
	/**< The relevant database object (cloned, is not persistent in
	 * the database itself).
	 */
	unsigned char	hash[NLPCR_HASHLEN];
	/**< The database key/hash value (not nil-terminated).
	 */
	int		fd;
	/**< Active file descriptor.  Is set to an open stream during
	 * the connection lifetime.
	 */
};

/**
 * Scan whether we fall within the LWS lexical token as defined by RFC
 * 2616 and 1945.  Note that this checks for individual elements, and
 * does not check for proper stream conformance (for instance, this will
 * take CR SP LF, which shouldn't happen).
 *
 * @param c The token to check.
 *
 * @return TRUE on success, FALSE on failure.
 */
static int
ishttp_lws(int c)
{
	switch (c) {
	case (13):
		/*FALLTHROUGH*/
	case (10):
		/*FALLTHROUGH*/
	case (32):
		/*FALLTHROUGH*/
	case (9):
		return(1);
	default:
		break;
	}
	return(0);
}

/**
 * Match a header key and capture a pointer to the data field.  This
 * conforms with HTTP/1.1 and HTTP/1.0, RFC 2616 and 1945, respectively,
 * that define the syntax of a header line.
 *
 * @param p The input string (header-line).
 * @param sz The size of p (header-line).
 * @param match The header key to find.
 * @param res On success, will be set to the beginning of the
 * header-value in p.
 * @param ressz On success, will be set to the remaining space in p from
 * res (may be 0, if there is no header-value).
 *
 * @return 1 on match, 0 on no match, -1 on protocol violation.  Note
 * that this may return a zero-length result.
 */
static int
match_header_key(char *p, size_t sz, const char *match,
		char **res, size_t *ressz)
{
	size_t		c, i;

	assert(NULL != p);
	assert(NULL != match);
	assert(NULL != res);
	assert(NULL != ressz);

	/*
	 * XXX.  This entire area needs to be carefully examined for
	 * errors.  It does string handling on uncertain data, and
	 * although there is a great deal of bounds checking,
	 * something's likely to get through.
	 */

	if (sz < (c = strlen(match)))
		return(0);
	for (i = 0; i < c; i++)
		if (toupper(*p++) != toupper(match[i]))
			return(0);
	if (*p++ != ':') 
		/* XXX - protocol violation. */
		return(-1);
	for ( ; i < sz && ishttp_lws(*p); p++) 
		/* Skips LWS. */ ;
	*res = p;
	*ressz = sz - i;
	return(1);
}

/**
 * Matches an un-terminated string with whether a pragma no-cache has
 * been set.
 *
 * @param p The header-value.
 * @param psz The header-value size (may be 0 if there is no
 * header-value).
 *
 * @return 1 on success, 0 on failure (no match).
 */
static int
match_pragma_nocache(char *p, size_t psz)
{
	size_t		i, msz;
	char		m[] = "no-cache";

	if (psz < (msz = strlen(m)))
		return(0);
	for (i = 0; i < msz; i++)
		if (toupper(p[i] != toupper(m[i])))
			return(0);
	return(1);
}

/**
 * Matches an un-terminated string with the integral representation of
 * the cache's expiration time.
 *
 * @param p The header-value.
 * @param psz The header-value size (may be 0 if there is no
 * header-value).
 * @param exp On success, set to the expiration date.
 *
 * @return 1 on success, -1 on protocol error.
 */
static int
match_expires(char *p, size_t psz, uint64_t *exp)
{
	char tmp	[psz + 1];
	time_t		tm;

	assert(NULL != p);
	assert(NULL != exp);

	/* Copy into a nil-terminated string. */
	memset(tmp, 0, sizeof(tmp));
	memcpy(tmp, p, psz);

	if (INT32_MAX == (tm = curl_getdate(tmp, NULL)) ||
			-1 == tm)
		/* XXX - protocol violation. */
		return(-1);

	*exp = (uint64_t)tm;
	return(1);
}

/**
 * Matches an un-terminated string with the integral representation of
 * media type (text/xml, text/html, ...) and charset (UTF-8, ...) (if
 * possible).
 *
 * @param p The header-value.
 * @param psz The header-value size (may be 0 if there is no
 * header-value).
 * @param type On success, set to the integral media type.
 * @param charset On success, set to the beginning of the character-set
 * field in p (header-value).
 * @param charsetssz On success, set to the length of charset.
 *
 * @return 1 on success, 0 on failure (no media-type match), -1 on
 * protocol error.
 */
static int
match_content_type(char *p, size_t psz, uint8_t *type,
		char **charset, size_t *charsetsz)
{
	size_t		j, i, sz;
	char 		*pp, m[] = "charset=";

	assert(NULL != p);
	assert(NULL != type);
	assert(NULL != charset);
	assert(NULL != charsetsz);

	/*
	 * XXX.  This entire area needs to be carefully examined for
	 * errors.  It does string handling on uncertain data, and
	 * although there is a great deal of bounds checking,
	 * something's likely to get through.
	 */

	if (psz == 0)
		/* XXX - protocol error. */
		return(0);

#define CONTENT_TYPE(str, val) \
	if ((sz = strlen((str))) < psz) { \
		pp = p; \
		for (j = 0, i = 0; i < sz; i++, j++) \
			if (toupper(pp[j]) != toupper((str)[i])) \
				break; \
		if (i == sz) { \
			*type = (val); \
			goto found; \
		} \
	}

	CONTENT_TYPE("application/xhtml+html", NLPCDBR_TYPE_APP_XHTML);
	CONTENT_TYPE("text/html", NLPCDBR_TYPE_TXT_HTML);
	CONTENT_TYPE("text/xml", NLPCDBR_TYPE_TXT_XML);
	CONTENT_TYPE("app/xml", NLPCDBR_TYPE_APP_XML);

	/* Exit strategy if no media-type found. */
	return(0);

found:	/* Media-type found.  Now parse charset. */

	*charset = NULL;
	if (pp[j++] != ';') 
		return(1);

	for ( ; j < psz && ishttp_lws(pp[j]); j++) 
		/* Skips LWS. */ ;
	if (j == psz)
		/* XXX - protocol error. */
		return(-1);

	/* Now skip through the "charset=" part. */
	sz = strlen(m);
	for (i = 0; i < sz; i++, j++)
		if (toupper(pp[j]) != toupper(m[i]))
			return(0);

	*charset = &pp[j];
	i = j;
	for ( ; j < psz && ! ishttp_lws(pp[j]); j++)
		/* Skip to LWS. */ ;
	*charsetsz = j - i;
	return(1);
}

/*anon*/ size_t
curl_data(void *ptr, size_t smem, size_t nmem, void *arg)
{
	ssize_t		rc;
	struct ecurl 	*ec;

	assert(NULL != arg);
	assert(NULL != ptr);
	assert(sizeof(char) == smem);
	assert((smem * nmem) < SIZE_MAX);

	ec = (struct ecurl *)arg;

	if (-1 == (rc = write(ec->fd, ptr, nmem)))
		err(EXIT_FAILURE, "write: %s", ec->rec.cache);
	if (rc != (ssize_t)nmem) 
		errx(EXIT_FAILURE, "write (underrun): %s", 
				ec->rec.cache);

	return(nmem);
}

/*anon*/ size_t
curl_header(void *ptr, size_t smem, size_t nmem, void *arg)
{
	struct ecurl 	*ec;
	size_t		psz, charsetsz, i;
	uint8_t		type;
	uint64_t	exp;
	int		rc;
	char		*p, *charset;

	assert(NULL != arg);
	assert(NULL != ptr);
	assert(sizeof(char) == smem);
	assert((smem * nmem) < SIZE_MAX);

	ec = (struct ecurl *)arg;

	/* 
	 * First match the content-type, which optionally (HTTP/1.1)
	 * contains the character encoding.
	 */

	rc = match_header_key((char *)ptr, nmem, "content-type", 
			&p, &psz);
	if (1 == rc) {
		rc = match_content_type(p, psz, &type, &charset,
				&charsetsz);
		if (-1 == rc)
			/* XXX - protocol violation. */
			return(nmem + 1);

		ec->rec.type = type;
		if (NULL != charset) {
			if (NLPCR_CHARSETLEN - 1 <= charsetsz)
				/* XXX - protocol violation. */
				return(nmem + 1);
			memset(ec->rec.charset, 0, NLPCR_CHARSETLEN);
			for (i = 0; i < charsetsz; i++)
				ec->rec.charset[i] = 
					(char)tolower(charset[i]);
		}
		return(nmem);
	} else if (-1 == rc)
		/* XXX - protocol violation. */
		return(nmem + 1);

	/*
	 * Next try to parse the expiration, which is in both HTTP/1.0
	 * and 1.1
	 */

	rc = match_header_key((char *)ptr, nmem, "expires", &p, &psz);
	if (1 == rc) {
		if (-1 == (rc = match_expires(p, psz, &exp)))
			/* XXX - protocol violation. */
			return(nmem + 1);
		ec->rec.expires = exp;
		return(nmem);
	} else if (-1 == rc)
		/* XXX - protocol violation. */
		return(nmem + 1);

	/*
	 * See if we're set not to cache (HTTP/1.0).
	 */

	rc = match_header_key((char *)ptr, nmem, "pragma", &p, &psz);
	if (1 == rc) {
		if (match_pragma_nocache(p, psz))
			ec->rec.expires = UINT64_MAX;
		return(nmem);
	} else if (-1 == rc)
		/* XXX - protocol violation. */
		return(nmem + 1);

	/* 
	 * TODO: process Content-cache HTTP/1.1 directive.
	 */

	return(nmem);
}

/**
 * Safely execute curl_easy_setopt (with a long as argument). 
 */
static void
curl_optl(CURL *ec, CURLoption o, long p)
{
	CURLcode	cc;
	assert(NULL != ec);
	if (CURLE_OK != (cc = curl_easy_setopt(ec, o, p)))
		errx(EXIT_FAILURE, "curl_easy_setopt: %s", 
				curl_easy_strerror(cc));
}

/**
 * Safely execute curl_easy_setopt (with a pointer as argument). 
 */
static void
curl_optp(CURL *ec, CURLoption o, void *p)
{
	CURLcode	cc;
	assert(NULL != ec);
	assert(NULL != p);
	if (CURLE_OK != (cc = curl_easy_setopt(ec, o, p)))
		errx(EXIT_FAILURE, "curl_easy_setopt: %s", 
				curl_easy_strerror(cc));
}

/**
 * Safely add an element to the header list.
 */
static void
curl_head(struct curl_slist **h, const char *data)
{
	assert(NULL != h);
	assert(NULL != data);
	if (NULL == (*h = curl_slist_append(*h, data)))
		errx(EXIT_FAILURE, "curl_slist_append");
	nlpcdb_ddbg("using http header: %s", data);
}

/**
 * Safely execute curl_multi_remove_handle.
 */
static void
curl_rem(CURLM *mc, CURL *ec)
{
	CURLMcode	cm;
	assert(NULL != mc);
	assert(NULL != ec);
	if (CURLM_OK != (cm = curl_multi_remove_handle(mc, ec)))
		errx(EXIT_FAILURE, "curl_multi_remove_handle: %s", 
				curl_multi_strerror(cm));
}

/**
 * Safely execute curl_multi_add_handle.
 */
static void
curl_add(CURLM *mc, CURL *ec)
{
	CURLMcode	cm;
	assert(NULL != mc);
	assert(NULL != ec);
	if (CURLM_OK != (cm = curl_multi_add_handle(mc, ec)))
		errx(EXIT_FAILURE, "curl_multi_add_handle: %s", 
				curl_multi_strerror(cm));
}

static void
init_lang(const char *lang, char header[256])
{
	int		rc;
	memset(header, 0, 256);
	rc = NULL == lang ?
		snprintf(header, 255, "Accept-Language: *") :
		snprintf(header, 255, "Accept-Language: %s, *", lang);
	if (255 < rc)
		errx(EXIT_FAILURE, "string too long: --lang");
}

/**
 * Initialise the curl subsystem and construct the initial pool of
 * connection objects (zeroed).
 */
static void
curl_init(CURLM **mcurl, struct ecurl **ecurls, size_t ecurlsz,
		struct curl_slist **h, const char *lang)
{
	CURLcode 	cc;

	assert(NULL != mcurl);
	assert(NULL != h);
	assert(NULL != ecurls);
	assert(0 < ecurlsz);

	/* Global initialisation first. */

	if (CURLE_OK != (cc = curl_global_init(CURL_GLOBAL_ALL)))
		errx(EXIT_FAILURE, "curl_global_init: %s",
				curl_easy_strerror(cc));

	/* Multi-connection initialisation follows. */

	if (NULL == (*mcurl = curl_multi_init())) 
		errx(EXIT_FAILURE, "curl_multi_init");

	/* Lastly, allocate maximum connection pool. */

	assert(ecurlsz > 0);
	if (NULL == (*ecurls = calloc(ecurlsz, sizeof(struct ecurl))))
		err(EXIT_FAILURE, "calloc: %zd, %zd", ecurlsz, 
				sizeof(struct ecurl));

	/*
	 * Headers.  This may be changed from application to
	 * application.  We want to emphasis xml-based formats because
	 * they're easier to parse (xhtml).  Also, let's suggest UTF-8
	 * because we'll need to do the conversation work regardless.
	 */

	curl_head(h, "User-Agent: nlpcrawl");
	curl_head(h, "Accept: application/xhtml+html,"
			"text/html,text/xml,application/xml");
	curl_head(h, "Accept-Charset: utf-8;qs=1,*");
	curl_head(h, lang);
}

/**
 * Cleanup the curl subsystem and connection pool.
 */
static void
curl_clean(CURLM *mcurl, struct ecurl *ecurls, size_t ecurlsz,
		struct curl_slist *headers)
{
	CURLMcode 	mc;
	size_t 		i;

	assert(ecurlsz > 0);
	if (NULL != mcurl && NULL != ecurls)
		for (i = 0; i < ecurlsz; i++)
			if (NULL != ecurls[i].curl) {
				curl_rem(mcurl, ecurls[i].curl);
				curl_easy_cleanup(ecurls[i].curl);
				ecurls[i].curl = NULL;
			}
	if (NULL != ecurls)
		free(ecurls);
	if (NULL != mcurl && CURLM_OK != 
			(mc = curl_multi_cleanup(mcurl))) 
		errx(EXIT_FAILURE, "curl_multi_cleanup: %s",
				curl_multi_strerror(mc));

	if (NULL != headers)
		curl_slist_free_all(headers);
	curl_global_cleanup();
}

/**
 * Schedule the next uri to fetch.  This fully initialisees the
 * connection object.  URIs are groked from the database if their
 * respective state is NLPCDBR_DIRTY or NLPCDBR_NEW. See
 * nlpcdb_nextmatch for return values.
 *
 * FIXME: this should be using a database transaction in order to
 * guarantee that multiple applications scanning the database (and
 * changing values) will not collide on a single entity.
 */
static int
sched_next(const char *file, CURLM *mcurl, struct ecurl *p, 
		unsigned long burst, struct curl_slist *h)
{
	unsigned char 	hash[NLPCR_HASHLEN];
	int		rc, flags;
	mode_t		mode;
	struct nlpcdbr 	rec, match;
	uint8_t 	fields;

	assert(NULL != mcurl);
	assert(NULL != file);
	assert(NULL != p);
	assert(NULL != h);

	mode = 0644;
	flags = O_WRONLY | O_TRUNC | O_CREAT;
	fields = NLPCDBR_FIELD_STATE;
	match.state = NLPCDBR_NEW | NLPCDBR_DIRTY;

	/*
	 * Fetch the next available entity to find.
	 */

	if (1 != (rc = nlpcdb_nextmatch(file, hash, &rec, &match, 
					fields, burst)))
		return(rc);

	if (NULL == p->curl && NULL == (p->curl = curl_easy_init()))
		errx(EXIT_FAILURE, "curl_easy_init");

	memcpy(&p->rec, &rec, sizeof(struct nlpcdbr));
	memcpy(p->hash, hash, NLPCR_HASHLEN);

	/*
	 * Set the appropriate libcurl callbacks.
	 */

	curl_add(mcurl, p->curl);
	/* TODO: URL-encode. */
	curl_optp(p->curl, CURLOPT_URL, p->rec.uri);
	curl_optp(p->curl, CURLOPT_HEADERFUNCTION, curl_header);
	curl_optp(p->curl, CURLOPT_WRITEFUNCTION, curl_data);
	curl_optp(p->curl, CURLOPT_WRITEDATA, p);
	curl_optp(p->curl, CURLOPT_HEADERDATA, p);
	/* Now HTTP options: */
	curl_optl(p->curl, CURLOPT_MAXREDIRS, 32);
	curl_optl(p->curl, CURLOPT_FOLLOWLOCATION, 1);

	/* 
	 * TODO: set the If-Modified-Since header for each
	 * already-fetched page.
	 */

	curl_optp(p->curl, CURLOPT_HTTPHEADER, h);

	/* 
	 * Configure the transfer environment. 
	 */

	if (-1 == (p->fd = open(p->rec.cache, flags, mode)))
		err(EXIT_FAILURE, "open: %s", p->rec.cache);
	/* Zero out fetched parameters. */
	p->rec.expires = 0;
	p->rec.type = NLPCDBR_TYPE_NONE;
	memset(p->rec.charset, 0, NLPCR_CHARSETLEN);

	nlpcdb_ddbg("scheduled: %s", p->rec.uri);

	return(1);
}

static void
postscan(const char *file, CURLM *mcurl, struct ecurl *ec,
		CURLMsg *cmsg)
{
	time_t		tm;

	assert(NULL != file);
	assert(NULL != mcurl);
	assert(NULL != ec);
	assert(NULL != cmsg);

	/* 
	 * Regardless out state-change below, we must tip the access and
	 * modification times.  
	 */

	if ((time_t)-1 == (tm = time(NULL)))
		err(EXIT_FAILURE, "time");

	ec->rec.atime = (uint64_t)tm;
	ec->rec.mtime = (uint64_t)tm;

	/* 
	 * Now process the error code and its effects on our state.
	 * Temporary errors push us into UNREACH, while fatal errors
	 * push us into DEAD.  Success puts us into FRESH and the mtime
	 * is updated.  Also process internal errors (abort).
	 */
	
	/* 
	 * TODO: closer examination of error codes.
	 */

	switch(cmsg->data.result) {
	case (CURLE_OK):
		ec->rec.state = NLPCDBR_FRESH;
		nlpcdb_dbg("result: %s, NLPCDBR_FRESH (%s)", 
				ec->rec.uri,
			curl_easy_strerror(cmsg->data.result));
		break;
	case (CURLE_COULDNT_RESOLVE_PROXY):
		/*FALLTHROUGH*/
	case (CURLE_COULDNT_RESOLVE_HOST):
		/*FALLTHROUGH*/
	case (CURLE_COULDNT_CONNECT):
		/*FALLTHROUGH*/
	case (CURLE_OPERATION_TIMEDOUT):
		ec->rec.state = NLPCDBR_UNREACH;
		nlpcdb_dbg("result: %s, NLPCDBR_UNREACH (%s)", 
				ec->rec.uri,
			curl_easy_strerror(cmsg->data.result));
		break;
	case (CURLE_OUT_OF_MEMORY):
		/*FALLTHROUGH*/
	case (CURLE_ABORTED_BY_CALLBACK):
		/*FALLTHROUGH*/
	case (CURLE_BAD_FUNCTION_ARGUMENT):
		/*FALLTHROUGH*/
	case (CURLE_FILESIZE_EXCEEDED):
		errx(EXIT_FAILURE, "internal libcurl: %s", 
			curl_easy_strerror(cmsg->data.result));
		/*NOTREACHED*/
	default:
		ec->rec.state = NLPCDBR_DEAD;
		nlpcdb_dbg("result: %s, NLPCDBR_DEAD (%s)", 
				ec->rec.uri,
			curl_easy_strerror(cmsg->data.result));
		break;
	}
	
	/*
	 * Update the database entry, close out the active
	 * file-descriptor and remove the entry from the multicurl
	 * database.
	 */

	nlpcdb_update(file, ec->hash, &ec->rec);
	if (-1 == close(ec->fd))
		err(EXIT_FAILURE, "close: %s", ec->rec.cache);
	ec->fd = -1;
	curl_rem(mcurl, ec->curl);
}

/**
 * Scan entire database, downloading content in pools.
 */
static void
scan(const char *file, CURLM *mcurl, struct ecurl *ecurls, 
		size_t ecurlsz, size_t ecurlrsz, struct curl_slist *h)
{
	fd_set 		fds_read, fds_send, fds_expt;
	struct timeval	tm;
	size_t		i;
	int		mfds, run, lastrun, oninit;
	CURLMcode	cm;
	CURLMsg		*cmsg;

	assert(NULL != mcurl);
	assert(NULL != ecurls);
	assert(NULL != file);
	assert(NULL != h);
	assert(0 < ecurlrsz && ecurlrsz < INT_MAX);
	assert(0 < ecurlsz && ecurlsz <= ecurlrsz);

	nlpcdb_ddbg("scanning: %zd slots (%zd real)", 
			ecurlsz, ecurlrsz);

	lastrun = (int)ecurlsz;
	oninit = 1;

start:	/* Scan for completed transfers; iterate over dataset. */

	/* 
	 * First conditionally select on waiting data.  If the transfers
	 * have not yet been initialised then do not yet select.
	 *
	 * Note that we ignore curl's suggested timeout.
	 */

	timerclear(&tm);
	tm.tv_sec = 1;
	tm.tv_usec = 0;

	FD_ZERO(&fds_read);
	FD_ZERO(&fds_send);
	FD_ZERO(&fds_expt);

	if (CURLM_OK != (cm = curl_multi_fdset(mcurl, &fds_read, 
					&fds_send, &fds_expt, &mfds)))
		errx(EXIT_FAILURE, "curl_multi_fdset: %s",
				curl_multi_strerror(cm));

	/*
	 * Do not select if we've no file descriptors (transfer not yet
	 * opened: at [any handle's] initialisation). 
	 */

	if (oninit && -1 != mfds && -1 == select(mfds, &fds_read, 
				&fds_send, &fds_expt, &tm))
		err(EXIT_FAILURE, "select");
	oninit = 0;

	/* 
	 * Now actually read the data.  Once data has been acted on,
	 * check the queue of handles for completed transfers and
	 * perform accounting.
	 */

	for (cm = CURLM_CALL_MULTI_PERFORM; ; ) 
		if (CURLM_CALL_MULTI_PERFORM !=
				(cm = curl_multi_perform(mcurl, &run)))
			break;
	if (CURLM_OK != cm)
		errx(EXIT_FAILURE, "curl_multi_perform: %s",
				curl_multi_strerror(cm));

	assert(0 != lastrun);
	assert(run <= lastrun);

	while (NULL != (cmsg = curl_multi_info_read(mcurl, &run))) {
		assert(CURLMSG_DONE == cmsg->msg);

		/* 
		 * Look up the easy handle, then postprocess the
		 * completed (or failed) transfer.
		 * TODO - replace lookup with a hashtable.
		 */

		for (i = 0; i < ecurlrsz; i++)
			if (ecurls[i].curl == cmsg->easy_handle)
				break;
		assert(i < ecurlrsz);
		postscan(file, mcurl, &ecurls[i], cmsg);

		/* 
		 * Schedule the next available job.  TODO - use burst
		 * feature. 
		 */ 

		switch (sched_next(file, mcurl, &ecurls[i], 0, h)) {
		case (1):
			oninit = 1;
			break;
		case (-1):
			ecurlsz--;
			nlpcdb_ddbg("end-of-database (%zd pending)", 
					ecurlsz);
			curl_easy_cleanup(ecurls[i].curl);
			ecurls[i].curl = NULL;
			if (0 != ecurlsz)
				break;
			assert(0 == run);
			return;
		default:
			/* XXX - deprecate with bursting. */
			abort();
		}
	}

	/* Re-trigger the scanning process. */
	goto start;
}

/**
 * See the connection pool, resizing it if necessary (not enough
 * elements in the entire database).
 */
static size_t
preseed(const char *file, CURLM *mcurl, struct ecurl *ecurls, 
		size_t ecurlsz, struct curl_slist *h)
{
	size_t		i;

	assert(NULL != mcurl);
	assert(NULL != ecurls);
	assert(NULL != h);
	assert(0 < ecurlsz);

	for (i = 0; i < ecurlsz; i++)
		if (1 != sched_next(file, mcurl, &ecurls[i], 0, h))
			break;

	return(i);
}

static void
usage(void)
{
	extern char 	*__progname;
	fprintf(stderr, "usage: %s [-dv] [-f file] [-p integer] "
			"[--lang=lang]\n", __progname);
	exit(EXIT_FAILURE);
	/*NOTREACHED*/
}

static void
version(void)
{
	extern char	*__progname;
	fprintf(stdout, "%s version %s\n", __progname, VERSION);
	exit(EXIT_SUCCESS);
	/*NOTREACHED*/
}

/*main*/ int
main(int argc, char *argv[])
{
	int 		c;
	size_t 		ecurlsz, ecurlrsz;
	char 		*db, lang[256];
	CURLM 		*mcurl;
	struct ecurl	*ecurls;
	extern char 	*optarg;
	extern int	optind;
	struct curl_slist *head;
	static struct option longopts[] = {
		{ "lang", required_argument, NULL, 1 },
		{ NULL, 0, NULL, 0 }
	};

	/* Initialise members to defaults. */

	db = NLPC_DEFAULT_DB;
	ecurlrsz = NLPC_DEFAULT_COMMSZ;
	init_lang(NULL, lang);

	/* Sanitise. */

	assert(NULL != db);
	assert(ecurlrsz > 0);

	while (-1 != (c = getopt_long(argc, argv, "dvf:p:", 
					longopts, NULL))) 
		switch (c) {
		case ('v'):
			version();
			/*NOTREACHED*/
		case ('d'):
			nlpcdb_dbg_lvl++;
			break;
		case ('f'):
			db = optarg;
			break;
		case ('p'):
			if (0 > atoi(optarg))
				errx(EXIT_FAILURE, "integer out of "
						"range: -p");
			ecurlrsz = (size_t)atoi(optarg);
			break;
		/* Long options follow. */
		case (1):
			init_lang(optarg, lang);
			break;
		default:
			usage();
			/*NOTREACHED*/
		}
	argc -= optind;
	argv += optind;

	nlpcdb_setup(db);

	mcurl = NULL;
	ecurls = NULL;
	head = NULL;

	curl_init(&mcurl, &ecurls, ecurlrsz, &head, lang);

	if (0 != (ecurlsz = preseed(db, mcurl, ecurls, 
					ecurlrsz, head)))
		scan(db, mcurl, ecurls, ecurlsz, ecurlrsz, head);

	curl_clean(mcurl, ecurls, ecurlrsz, head);
	nlpcdb_close(db);

	return(EXIT_SUCCESS);
}
