/* $Id: list.c,v 1.4 2007-03-23 11:42:33 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>

#include <assert.h>
#include <err.h>
#include <getopt.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "nlpcrawl.h"

static char *
hash2str(const unsigned char hash[NLPCR_HASHLEN])
{
	int		i;
	static char	hashstr[(NLPCR_HASHLEN * 2) + 1];

	for (i = 0; i < NLPCR_HASHLEN; i++) 
		snprintf(&hashstr[i * 2], 3, "%.2x", hash[i]);
	hashstr[NLPCR_HASHLEN * 2] = 0;

	return(hashstr);
}

static char *
state2str(uint8_t state)
{
	static char str[32];
	memset(str, 0, 32);
	switch (state) {
	case (NLPCDBR_NEW):
		strncpy(str, "new", 32);
		break;
	case (NLPCDBR_FRESH):
		strncpy(str, "fresh", 32);
		break;
	case (NLPCDBR_DIRTY):
		strncpy(str, "dirty", 32);
		break;
	case (NLPCDBR_UNREACH):
		strncpy(str, "unreach", 32);
		break;
	case (NLPCDBR_DEAD):
		strncpy(str, "dead", 32);
		break;
	default:
		strncpy(str, "unknown", 32);
		break;
	}
	return(str);
}

static char *
type2str(uint8_t type)
{
	static char str[32];
	memset(str, 0, 32);
	switch (type) {
	case (NLPCDBR_TYPE_NONE):
		strncpy(str, "none", 32);
		break;
	case (NLPCDBR_TYPE_TXT_XML):
		strncpy(str, "txt/xml", 32);
		break;
	case (NLPCDBR_TYPE_APP_XML):
		strncpy(str, "app/xml", 32);
		break;
	case (NLPCDBR_TYPE_APP_XHTML):
		strncpy(str, "app/xhtml", 32);
		break;
	case (NLPCDBR_TYPE_TXT_HTML):
		strncpy(str, "txt/html", 32);
		break;
	default:
		strncpy(str, "unknown", 32);
		break;
	}
	return(str);
}

static void
show_long(unsigned long n, const struct nlpcdbr *rec, 
		const unsigned char hash[NLPCR_HASHLEN])
{
	time_t		atime, rtime, mtime, exp;

	atime = (time_t)rec->atime;
	mtime = (time_t)rec->mtime;
	rtime = (time_t)rec->ctime;
	exp = (time_t)rec->expires;

	printf("%-6lu%32s (%s)\n", n, hash2str(hash), 
			state2str(rec->state));
	printf("        uri:      %s\n", rec->uri);
	printf("        cache:    %s\n", rec->cache);
	printf("        content:  %s, %s\n", type2str(rec->type), 
			rec->charset[0] == 0 ? "none" : rec->charset);
	printf("        atime:    %s", atime == 0 ?
			"none\n" : ctime(&atime));
	printf("        ctime:    %s", rtime == 0 ?
			"none\n" : ctime(&rtime));
	printf("        mtime:    %s", mtime == 0 ?
			"none\n" : ctime(&mtime));
	printf("        expires:  %s", exp == 0 ?
			"none\n" : ctime(&exp));
}

static void
show(unsigned long n, const struct nlpcdbr *rec, 
		const unsigned char hash[NLPCR_HASHLEN])
{
	printf("%-6lu%32s%10s (%s; %s) %s\n", n, 
			hash2str(hash),
			state2str(rec->state), type2str(rec->type),
			rec->charset[0] == 0 ? "none" : rec->charset,
			rec->uri);
}

/**
 * Scan through the argument list and seed the database.  Ignore
 * duplicate entities.
 */
static void
list(int larg, const char *file)
{
	struct nlpcdbr 	rec;
	unsigned char	hash[NLPCR_HASHLEN];
	unsigned long	i;

	assert(NULL != file);

	i = 1;
	while (nlpcdb_next(file, hash, &rec))
		larg ? show_long(i++, &rec, hash) :
			show(i++, &rec, hash);
}

static void
usage(void)
{
	extern char 	*__progname;
	fprintf(stderr, "usage: %s [-vl] [-f file]\n", __progname);
	exit(EXIT_FAILURE);
	/*NOTREACHED*/
}

static void
version(void)
{
	extern char	*__progname;
	fprintf(stdout, "%s version %s\n", __progname, VERSION);
	exit(EXIT_SUCCESS);
	/*NOTREACHED*/
}

int
main(int argc, char *argv[])
{
	extern char 	*optarg;
	extern int 	optind;
	int 		c, larg;
	char 		*db;

	db = NLPC_DEFAULT_DB;
	larg = 0;

	while ((c = getopt(argc, argv, "vlf:")) != -1)
		switch (c) {
		case ('v'):
			version();
			/*NOTREACHED*/
		case ('l'):
			larg = 1;
			break;
		case ('f'):
			db = optarg;
			break;
		default:
			usage();
			/*NOTREACHED*/
		}
	argc -= optind;
	argv += optind;

	nlpcdb_setup(db);
	list(larg, db);
	nlpcdb_close(db);

	return(EXIT_SUCCESS);
}
