/* $Id: nlpcrawl.h,v 1.10 2007-03-22 15:33:12 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef NLPCRAWL_H
#define NLPCRAWL_H

__BEGIN_DECLS

/**
 * How many connections to maintain (default). 
 */
#define NLPC_DEFAULT_COMMSZ	(16)

/**
 * Location of nlpcrawl btree database.
 */
#define NLPC_DEFAULT_DB		"nlpcrawl.db"

/**
 * Default directory in which to store cached files.  This should be
 * configured to handle many files (one per URI).
 */
#define NLPC_DEFAULT_DIR	"/tmp"

/**
 * Length of un-terminated MD5 digest.
 */
#define NLPCR_HASHLEN		(16)

/** 
 * Length of cache file address (/tmp/path/to/cache/file).  Note that
 * this is the full pathname, not just the path.
 */
#define NLPCR_CACHELEN		(1024)

/**
 * Maximum URI length.  This is technically not supposed to be fixed.
 */
#define NLPCR_URILEN		(1024)

/**
 * Maximum charset length.  Grokked from ASN.1.
 */
#define NLPCR_CHARSETLEN	(64)

/**
 * This is documented in full in nlpcrawl(1).
 */
struct nlpcdbr {
	char uri[NLPCR_URILEN];
#define NLPCDBR_NEW		(1 << 0)
#define NLPCDBR_FRESH		(1 << 1)
#define NLPCDBR_DIRTY		(1 << 2)
#define NLPCDBR_UNREACH		(1 << 3)
#define NLPCDBR_DEAD		(1 << 4)
#define NLPCDBR_SCAN		(1 << 5)
#define NLPCDBR_PROC		(1 << 6)
	uint8_t state;
	uint64_t ctime;
	uint64_t mtime;
	uint64_t atime;
	char cache[NLPCR_CACHELEN];
#define NLPCDBR_TYPE_NONE	(1 << 0)
#define NLPCDBR_TYPE_TXT_XML	(1 << 1)
#define NLPCDBR_TYPE_TXT_HTML	(1 << 2)
#define NLPCDBR_TYPE_APP_XML	(1 << 3)
#define NLPCDBR_TYPE_APP_XHTML	(1 << 4)
	uint8_t type;
	char charset[NLPCR_CHARSETLEN];
	uint64_t expires;
};

/**
 * Construct the digest of a uri.
 */
extern void
nlpcdb_gethash(const char *uri, 
		unsigned char hash[NLPCR_HASHLEN]);

/**
 * Prepares the database.  Exits run-time on failure.
 */
extern void
nlpcdb_setup(const char *file);

/* 
 * Closes the database.  Ignores (but reports) errors.
 */
extern void
nlpcdb_close(const char *file);

/*
 * Add an object to the database.  Object is cloned on the heap before
 * addition, with members being appropriately re-written to NBO.
 *
 * @return TRUE on success, FALSE if record exists.
 */
extern int
nlpcdb_add(const char *file, const unsigned char hash[NLPCR_HASHLEN],
		const struct nlpcdbr *rec);

/**
 * Resets positional cursor.
 */
extern void
nlpcdb_reset(const char *file);

/**
 * Fetch the next record.  It's wise to call nlpcdb_reset before this
 * function, lest the current cursor be out of state.
 *
 * @return TRUE on success, FALSE if no more records.
 */
extern int
nlpcdb_next(const char *file, unsigned char hash[NLPCR_HASHLEN],
		struct nlpcdbr *rec);

#define NLPCDBR_FIELD_URI	(1 << 0)
#define NLPCDBR_FIELD_STATE	(1 << 1)
#define NLPCDBR_FIELD_CTIME	(1 << 2)
#define NLPCDBR_FIELD_MTIME	(1 << 3)
#define NLPCDBR_FIELD_ATIME	(1 << 4)

/**
 * Scan ahead (in bursts) for matches in the database.  If the burst
 * value is 0, then scans indefinitely.
 *
 * @return 1 with a match, 0 if burst has been reached, -1 if the
 * database has been exhausted.
 */
extern int
nlpcdb_nextmatch(const char *file, 
		unsigned char hash[NLPCR_HASHLEN],
		struct nlpcdbr *rec, const struct nlpcdbr *match,
		uint8_t fields, unsigned long burst);

/**
 * Update an existing record in the database.
 */
extern void
nlpcdb_update(const char *file, 
		unsigned char hash[NLPCR_HASHLEN],
		const struct nlpcdbr *rec);

/**
 * Write an informational message.  This is "level 1" debugging, and
 * should be used only for important messages.
 */
extern void
nlpcdb_dbg(const char *fmt, ...)
	__attribute__((format (printf, 1, 2)));

/**
 * Write a debugging message.  This is "level 2" debugging, which may be
 * used by really any message.
 */
extern void
nlpcdb_ddbg(const char *fmt, ...)
	__attribute__((format (printf, 1, 2)));

/** 
 * Debug message level.  1 is informational, 2 is verbose.  0 is no
 * debugging message (the default).
 */
extern int		nlpcdb_dbg_lvl;

__END_DECLS

#endif /*!NLPCRAWL_H*/
