/* $Id: scan.c,v 1.12 2007-03-23 11:42:33 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <assert.h>
#include <err.h>
#include <getopt.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "nlpcrawl.h"

static void
init(const char *cachedir, const char *uri, 
		unsigned char hash[NLPCR_HASHLEN], 
		struct nlpcdbr *rec)
{
	char		hashstr[(NLPCR_HASHLEN * 2) + 1];
	int		i;
	time_t		tm;

	if ((time_t)-1 == (tm = time(NULL)))
		err(EXIT_FAILURE, "time");

	assert(NULL != cachedir);
	assert(NULL != uri);
	assert(NULL != rec);

	memset(rec, 0, sizeof(struct nlpcdbr));
	memset(hash, 0, NLPCR_HASHLEN);

	strncpy(rec->uri, uri, NLPCR_URILEN - 1);

	nlpcdb_gethash(uri, hash);

	for (i = 0; i < NLPCR_HASHLEN; i++) 
		snprintf(&hashstr[i * 2], 3, "%.2x", hash[i]);
	hashstr[NLPCR_HASHLEN * 2] = 0;
	snprintf(rec->cache, NLPCR_CACHELEN - 1, "%s/%s",
			cachedir, hashstr);

	rec->state = NLPCDBR_NEW;
	rec->atime = rec->mtime = rec->ctime = (uint64_t)tm;
}

static void
seed(int dryrun, const char *db, const char *cachedir, int argc, 
		char *argv[])
{
	struct nlpcdbr 	rec;
	unsigned char	hash[NLPCR_HASHLEN];
	char		*p;

	assert(NULL != db);
	assert(NULL != cachedir);
	assert(NULL != argv);

	while (argc-- > 0 && NULL != (p = *argv++)) {
		if (strlen(p) > NLPCR_URILEN - 1) {
			warnx("skipping uri (too long): %.20s...", p);
			continue;
		}

		init(cachedir, p, hash, &rec);

		if ( ! dryrun && ! nlpcdb_add(db, hash, &rec))
			warnx("skipping uri (duplicate): %s", p);
		else
			nlpcdb_dbg("registered uri: %s", p);
	}
}

static int
process_xml(int dryrun, const char *file, const char *cachedir,
		const struct nlpcdbr *rec)
{
	assert(NULL != file);
	assert(NULL != cachedir);
	assert(NULL != rec);
	(void)dryrun;
	return(1);
}

static int
process_html(int dryrun, const char *file, const char *cachedir,
		const struct nlpcdbr *rec)
{
	assert(NULL != file);
	assert(NULL != cachedir);
	assert(NULL != rec);
	(void)dryrun;
	return(1);
}

/**
 * @return TRUE on success, FALSE on failure.
 */
static int
process(int dryrun, const char *file, const char *cachedir,
		const struct nlpcdbr *rec)
{
	assert(NULL != file);
	assert(NULL != cachedir);
	assert(NULL != rec);

	switch (rec->type) {
	case (NLPCDBR_TYPE_TXT_XML):
		/*FALLTHROUGH*/
	case (NLPCDBR_TYPE_APP_XML):
		/*FALLTHROUGH*/
	case (NLPCDBR_TYPE_APP_XHTML):
		return(process_xml(dryrun, file, cachedir, rec));
	case (NLPCDBR_TYPE_TXT_HTML):
		return(process_html(dryrun, file, cachedir, rec));
	default:
		break;
	}

	/* XXX - report filetype violation. */
	return(0);
}

static void
scan(int dryrun, const char *file, const char *cachedir)
{
	int		rc;
	uint8_t		fields;
	struct nlpcdbr	rec, match;
	unsigned char	hash[NLPCR_HASHLEN];
	time_t		tm;
	uint64_t	tms;

	assert(NULL != cachedir);
	assert(NULL != file);

	if ((time_t)-1 == (tm = time(NULL)))
		err(EXIT_FAILURE, "time");
	tms = (uint64_t)tm;

	match.state = NLPCDBR_FRESH;
	fields = NLPCDBR_FIELD_STATE;

	rc = nlpcdb_nextmatch(file, hash, &rec, &match,
			fields, 0);
	for ( ; rc == 1; rc = nlpcdb_nextmatch(file, hash, &rec, 
				&match, fields, 0)) {
		/* 
		 * Actual scan the contents from the file.  If the parse
		 * fails, mark the page as dead.
		 */
		rec.atime = tms;
		if (process(dryrun, file, cachedir, &rec)) {
			rec.state = NLPCDBR_SCAN;
			nlpcdb_dbg("scan complete: %s", rec.uri);
			if ( ! dryrun)
				nlpcdb_update(file, hash, &rec);
		} else {
			rec.state = NLPCDBR_DEAD;
			if ( ! dryrun)
				nlpcdb_update(file, hash, &rec);
		}
	}
	assert(0 != rc);
}

static void
usage(void)
{
	extern char 	*__progname;
	fprintf(stderr, "usage: %s [-vnd] [-f file] [-p dir] "
			"[uri ...]\n", __progname);
	exit(EXIT_FAILURE);
	/*NOTREACHED*/
}

static void
version(void)
{
	extern char	*__progname;
	fprintf(stdout, "%s version %s\n", __progname, VERSION);
	exit(EXIT_SUCCESS);
	/*NOTREACHED*/
}

int
main(int argc, char *argv[])
{
	int 		c, dryrun;
	char 		*db, *cachedir;
	extern char 	*optarg;
	extern int 	optind;

	db = NLPC_DEFAULT_DB;
	cachedir = NLPC_DEFAULT_DIR;
	dryrun = 0;

	while ((c = getopt(argc, argv, "nvdf:p:")) != -1)
		switch (c) {
		case ('v'):
			version();
			/*NOTREACHED*/
		case ('d'):
			nlpcdb_dbg_lvl++;
			break;
		case ('n'):
			dryrun = 1;
			break;
		case ('p'):
			cachedir = optarg;
			break;
		case ('f'):
			db = optarg;
			break;
		default:
			usage();
			/*NOTREACHED*/
		}
	argc -= optind;
	argv += optind;

	if (strlen(cachedir) + NLPCR_HASHLEN + 2 >= 
			(NLPCR_CACHELEN - 1))
		errx(EXIT_FAILURE, "directory too long: -p %s", 
				cachedir);

	nlpcdb_setup(db);

	seed(dryrun, db, cachedir, argc, argv);
	scan(dryrun, db, cachedir);

	nlpcdb_close(db);

	return(EXIT_SUCCESS);
}
