	/* $Id: util.c,v 1.10 2007-03-22 14:35:36 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <arpa/inet.h>
#include <sys/types.h>

#include <assert.h>
#include <err.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <db.h>

#include "nlpcrawl.h"

static DB 		*dbp = NULL;
static DBC		*dbcursep = NULL;
/*extern*/ int		nlpcdb_dbg_lvl = 0;

#define nlpc_ntohll(x) \
	(((uint64_t)ntohl(x) << 32) + ntohl(x >> 32))
#define nlpc_htonll(x) \
	(((uint64_t)htonl((x)) << 32) + htonl((x) >> 32))

/*-------------------------------------------------------------------*/

typedef struct MD5Context {
	uint32_t buf[4];
	uint32_t bits[2];
	unsigned char in[64];
} MD5_CTX;

/*
 * This code implements the MD5 message-digest algorithm.  The algorithm
 * is due to Ron Rivest.  This code was written by Colin Plumb in 1993,
 * no copyright is claimed.  This code is in the public domain; do with
 * it what you wish.
 *
 * Equivalent code is available from RSA Data Security, Inc.  This code
 * has been tested against that, and is equivalent, except that you
 * don't need to include two pages of legalese with every copy.
 *
 * To compute the message digest of a chunk of bytes, declare an
 * MD5Context structure, pass it to MD5Init, call MD5Update as needed on
 * buffers full of bytes, and then call MD5Final, which will fill a
 * supplied 16-byte array with the digest.
 */

/* This code was modified in 1997 by Jim Kingdon of Cyclic Software to
   not require an integer type which is exactly 32 bits.  This work
   draws on the changes for the same purpose by Tatu Ylonen
   <ylo@cs.hut.fi> as part of SSH, but since I didn't actually use that
   code, there is no copyright issue.  I hereby disclaim copyright in
   any changes I have made; this code remains in the public domain.  */

/* Little-endian byte-swapping routines.  Note that these do not
   depend on the size of datatypes such as uint32, nor do they require
   us to detect the endianness of the machine we are running on.  It is
   possible they should be macros for speed, but I would be surprised if
   they were a performance bottleneck for MD5.  */

static uint32_t
getu32(const unsigned char *addr)
{
	return (((((uint32_t)addr[3] << 8) | addr[2]) << 8)
		| addr[1]) << 8 | addr[0];
}

static void
putu32(uint32_t data, unsigned char *addr)
{
	addr[0] = (unsigned char)data;
	addr[1] = (unsigned char)(data >> 8);
	addr[2] = (unsigned char)(data >> 16);
	addr[3] = (unsigned char)(data >> 24);
}

/* The four core functions - F1 is optimized somewhat */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (x ^ y ^ z)
#define F4(x, y, z) (y ^ (x | ~z))

/* This is the central step in the MD5 algorithm. */
#define MD5STEP(f, w, x, y, z, data, s) \
	( w += f(x, y, z) + data, w &= 0xffffffff, \
	  w = w<<s | w>>(32-s), w += x )

/*
 * The core of the MD5 algorithm, this alters an existing MD5 hash to
 * reflect the addition of 16 longwords of new data.  MD5Update blocks
 * the data and converts bytes into longwords for this routine.
 */
static void
MD5Transform(uint32_t buf[4], const unsigned char inraw[64])
{
	register uint32_t a, b, c, d;
	uint32_t in[16];
	int i;

	for (i = 0; i < 16; ++i)
		in[i] = getu32 (inraw + 4 * i);

	a = buf[0];
	b = buf[1];
	c = buf[2];
	d = buf[3];

	MD5STEP(F1, a, b, c, d, in[ 0]+0xd76aa478,  7);
	MD5STEP(F1, d, a, b, c, in[ 1]+0xe8c7b756, 12);
	MD5STEP(F1, c, d, a, b, in[ 2]+0x242070db, 17);
	MD5STEP(F1, b, c, d, a, in[ 3]+0xc1bdceee, 22);
	MD5STEP(F1, a, b, c, d, in[ 4]+0xf57c0faf,  7);
	MD5STEP(F1, d, a, b, c, in[ 5]+0x4787c62a, 12);
	MD5STEP(F1, c, d, a, b, in[ 6]+0xa8304613, 17);
	MD5STEP(F1, b, c, d, a, in[ 7]+0xfd469501, 22);
	MD5STEP(F1, a, b, c, d, in[ 8]+0x698098d8,  7);
	MD5STEP(F1, d, a, b, c, in[ 9]+0x8b44f7af, 12);
	MD5STEP(F1, c, d, a, b, in[10]+0xffff5bb1, 17);
	MD5STEP(F1, b, c, d, a, in[11]+0x895cd7be, 22);
	MD5STEP(F1, a, b, c, d, in[12]+0x6b901122,  7);
	MD5STEP(F1, d, a, b, c, in[13]+0xfd987193, 12);
	MD5STEP(F1, c, d, a, b, in[14]+0xa679438e, 17);
	MD5STEP(F1, b, c, d, a, in[15]+0x49b40821, 22);

	MD5STEP(F2, a, b, c, d, in[ 1]+0xf61e2562,  5);
	MD5STEP(F2, d, a, b, c, in[ 6]+0xc040b340,  9);
	MD5STEP(F2, c, d, a, b, in[11]+0x265e5a51, 14);
	MD5STEP(F2, b, c, d, a, in[ 0]+0xe9b6c7aa, 20);
	MD5STEP(F2, a, b, c, d, in[ 5]+0xd62f105d,  5);
	MD5STEP(F2, d, a, b, c, in[10]+0x02441453,  9);
	MD5STEP(F2, c, d, a, b, in[15]+0xd8a1e681, 14);
	MD5STEP(F2, b, c, d, a, in[ 4]+0xe7d3fbc8, 20);
	MD5STEP(F2, a, b, c, d, in[ 9]+0x21e1cde6,  5);
	MD5STEP(F2, d, a, b, c, in[14]+0xc33707d6,  9);
	MD5STEP(F2, c, d, a, b, in[ 3]+0xf4d50d87, 14);
	MD5STEP(F2, b, c, d, a, in[ 8]+0x455a14ed, 20);
	MD5STEP(F2, a, b, c, d, in[13]+0xa9e3e905,  5);
	MD5STEP(F2, d, a, b, c, in[ 2]+0xfcefa3f8,  9);
	MD5STEP(F2, c, d, a, b, in[ 7]+0x676f02d9, 14);
	MD5STEP(F2, b, c, d, a, in[12]+0x8d2a4c8a, 20);

	MD5STEP(F3, a, b, c, d, in[ 5]+0xfffa3942,  4);
	MD5STEP(F3, d, a, b, c, in[ 8]+0x8771f681, 11);
	MD5STEP(F3, c, d, a, b, in[11]+0x6d9d6122, 16);
	MD5STEP(F3, b, c, d, a, in[14]+0xfde5380c, 23);
	MD5STEP(F3, a, b, c, d, in[ 1]+0xa4beea44,  4);
	MD5STEP(F3, d, a, b, c, in[ 4]+0x4bdecfa9, 11);
	MD5STEP(F3, c, d, a, b, in[ 7]+0xf6bb4b60, 16);
	MD5STEP(F3, b, c, d, a, in[10]+0xbebfbc70, 23);
	MD5STEP(F3, a, b, c, d, in[13]+0x289b7ec6,  4);
	MD5STEP(F3, d, a, b, c, in[ 0]+0xeaa127fa, 11);
	MD5STEP(F3, c, d, a, b, in[ 3]+0xd4ef3085, 16);
	MD5STEP(F3, b, c, d, a, in[ 6]+0x04881d05, 23);
	MD5STEP(F3, a, b, c, d, in[ 9]+0xd9d4d039,  4);
	MD5STEP(F3, d, a, b, c, in[12]+0xe6db99e5, 11);
	MD5STEP(F3, c, d, a, b, in[15]+0x1fa27cf8, 16);
	MD5STEP(F3, b, c, d, a, in[ 2]+0xc4ac5665, 23);

	MD5STEP(F4, a, b, c, d, in[ 0]+0xf4292244,  6);
	MD5STEP(F4, d, a, b, c, in[ 7]+0x432aff97, 10);
	MD5STEP(F4, c, d, a, b, in[14]+0xab9423a7, 15);
	MD5STEP(F4, b, c, d, a, in[ 5]+0xfc93a039, 21);
	MD5STEP(F4, a, b, c, d, in[12]+0x655b59c3,  6);
	MD5STEP(F4, d, a, b, c, in[ 3]+0x8f0ccc92, 10);
	MD5STEP(F4, c, d, a, b, in[10]+0xffeff47d, 15);
	MD5STEP(F4, b, c, d, a, in[ 1]+0x85845dd1, 21);
	MD5STEP(F4, a, b, c, d, in[ 8]+0x6fa87e4f,  6);
	MD5STEP(F4, d, a, b, c, in[15]+0xfe2ce6e0, 10);
	MD5STEP(F4, c, d, a, b, in[ 6]+0xa3014314, 15);
	MD5STEP(F4, b, c, d, a, in[13]+0x4e0811a1, 21);
	MD5STEP(F4, a, b, c, d, in[ 4]+0xf7537e82,  6);
	MD5STEP(F4, d, a, b, c, in[11]+0xbd3af235, 10);
	MD5STEP(F4, c, d, a, b, in[ 2]+0x2ad7d2bb, 15);
	MD5STEP(F4, b, c, d, a, in[ 9]+0xeb86d391, 21);

	buf[0] += a;
	buf[1] += b;
	buf[2] += c;
	buf[3] += d;
}

/*
 * Start MD5 accumulation.  Set bit count to 0 and buffer to mysterious
 * initialization constants.
 */
static void
MD5Init(struct MD5Context *ctx)
{
	ctx->buf[0] = 0x67452301;
	ctx->buf[1] = 0xefcdab89;
	ctx->buf[2] = 0x98badcfe;
	ctx->buf[3] = 0x10325476;

	ctx->bits[0] = 0;
	ctx->bits[1] = 0;
}

/*
 * Update context to reflect the concatenation of another buffer full
 * of bytes.
 */
static void
MD5Update(struct MD5Context *ctx, unsigned char const *buf, unsigned len)
{
	uint32_t t;

	/* Update bitcount */

	t = ctx->bits[0];
	if ((ctx->bits[0] = (t + ((uint32_t)len << 3)) & 
				0xffffffff) < t)
		ctx->bits[1]++;	/* Carry from low to high */
	ctx->bits[1] += len >> 29;

	t = (t >> 3) & 0x3f;	/* Bytes already in shsInfo->data */

	/* Handle any leading odd-sized chunks */

	if ( t ) {
		unsigned char *p = ctx->in + t;

		t = 64-t;
		if (len < t) {
			memcpy(p, buf, len);
			return;
		}
		memcpy(p, buf, t);
		MD5Transform(ctx->buf, ctx->in);
		buf += t;
		len -= t;
	}

	/* Process data in 64-byte chunks */

	while (len >= 64) {
		memcpy(ctx->in, buf, 64);
		MD5Transform(ctx->buf, ctx->in);
		buf += 64;
		len -= 64;
	}

	/* Handle any remaining bytes of data. */

	memcpy(ctx->in, buf, len);
}

/*
 * Final wrapup - pad to 64-byte boundary with the bit pattern 
 * 1 0* (64-bit count of bits processed, MSB-first)
 */
static void
MD5Final(unsigned char digest[16], struct MD5Context *ctx)
{
	unsigned count;
	unsigned char *p;

	/* Compute number of bytes mod 64 */
	count = (ctx->bits[0] >> 3) & 0x3F;

	/* Set the first char of padding to 0x80.  This is safe since
	 * there is always at least one byte free */
	p = ctx->in + count;
	*p++ = 0x80;

	/* Bytes of padding needed to make 64 bytes */
	count = 64 - 1 - count;

	/* Pad out to 56 mod 64 */
	if (count < 8) {
		/* Two lots of padding:  Pad the first block to 64 bytes
		 * */
		memset(p, 0, count);
		MD5Transform(ctx->buf, ctx->in);

		/* Now fill the next block with 56 bytes */
		memset(ctx->in, 0, 56);
	} else {
		/* Pad block to 56 bytes */
		memset(p, 0, count-8);
	}

	/* Append length in bits and transform */
	putu32(ctx->bits[0], ctx->in + 56);
	putu32(ctx->bits[1], ctx->in + 60);

	MD5Transform(ctx->buf, ctx->in);
	putu32(ctx->buf[0], digest);
	putu32(ctx->buf[1], digest + 4);
	putu32(ctx->buf[2], digest + 8);
	putu32(ctx->buf[3], digest + 12);
	memset(ctx, 0, sizeof(ctx));	/* In case it's sensitive */
}

/*-------------------------------------------------------------------*/

static void
nlpcdbr_copyin(struct nlpcdbr *dst, const struct nlpcdbr *src)
{
	assert(NULL != dst);
	assert(NULL != src);

	memcpy(dst->uri, src->uri, NLPCR_URILEN);
	memcpy(dst->cache, src->cache, NLPCR_CACHELEN);
	memcpy(dst->charset, src->charset, NLPCR_CHARSETLEN);
	dst->state = src->state;
	dst->type = src->type;
	dst->expires = nlpc_htonll(src->expires);
	dst->ctime = nlpc_htonll(src->ctime);
	dst->atime = nlpc_htonll(src->atime);
	dst->mtime = nlpc_htonll(src->mtime);
}

static void
nlpcdbr_copyout(struct nlpcdbr *dst, const struct nlpcdbr *src)
{
	assert(NULL != dst);
	assert(NULL != src);

	memcpy(dst->uri, src->uri, NLPCR_URILEN);
	memcpy(dst->cache, src->cache, NLPCR_CACHELEN);
	memcpy(dst->charset, src->charset, NLPCR_CHARSETLEN);

	dst->state = src->state;
	dst->type = src->type;
	dst->expires = nlpc_ntohll(src->expires);
	dst->ctime = nlpc_ntohll(src->ctime);
	dst->atime = nlpc_ntohll(src->atime);
	dst->mtime = nlpc_ntohll(src->mtime);
}

static void
nlpcdb_cursorclose(const char *file)
{
	int rc;

	if (NULL == dbcursep)
		return;

	assert(NULL != file);
	if (0 != (rc = dbcursep->c_close(dbcursep)))
		errx(EXIT_FAILURE, "database cursor: %s (%s)",
				file, db_strerror(rc));
	dbcursep = NULL;
}

static void
nlpcdb_cursorreset(const char *file)
{
	int rc;

	assert(NULL != file);
	assert(NULL != dbp);

	nlpcdb_cursorclose(file);

	if (0 != (rc = dbp->cursor(dbp, NULL, &dbcursep, 0)))
		errx(EXIT_FAILURE, "database cursor: %s (%s)",
				file, db_strerror(rc));
}

static void
nlpcdb_cursorinit(const char *file)
{
	if (NULL != dbcursep) 
		return;
	nlpcdb_cursorreset(file);
}

static int
nlpcdbr_nextraw(const char *file, unsigned char **hash,
		struct nlpcdbr **rec)
{
	int rc;
	DBT key, data;

	assert(NULL != file);
	assert(NULL != dbp);
	assert(NULL != rec);

	nlpcdb_cursorinit(file);

	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	assert(NULL != dbcursep);
	if (0 != (rc = dbcursep->c_get(dbcursep, &key, 
					&data, DB_NEXT))) {
		if (DB_NOTFOUND == rc)
			return(0);
		errx(EXIT_FAILURE, "database scan: %s (%s)",
				file, db_strerror(rc));
	}

	assert(16 == key.size);
	*hash = (unsigned char *)key.data;

	assert(sizeof(struct nlpcdbr) == data.size);
	*rec = (struct nlpcdbr *)data.data;

	return(1);
}

/*extern*/ void
nlpcdb_setup(const char *file)
{
	int rc;

	assert(file != NULL);
	assert(NULL == dbp);

	if (0 != db_create(&dbp, NULL, 0))
		errx(EXIT_FAILURE, "dbcreate");
	if (0 == (rc = dbp->open(dbp, NULL, file, NULL, DB_BTREE, 
					DB_CREATE, 0644)))
		return;

	nlpcdb_close(file);
	errx(EXIT_FAILURE, "database open: %s (%s)", file,
			db_strerror(rc));
}

/*extern*/ void
nlpcdb_close(const char *file)
{
	int rc;

	assert(file != NULL);

	if (NULL == dbp) {
		assert(NULL == dbcursep);
		return;
	}
	if (0 == (rc = dbp->close(dbp, 0)))
		return;

	nlpcdb_cursorclose(file);
	dbp = NULL;
	warn("database close: %s (%s)", file, db_strerror(rc));
}

/*extern*/ void
nlpcdb_reset(const char *file)
{
	assert(NULL != file);
	nlpcdb_cursorreset(file);
}

/*extern*/ int
nlpcdb_next(const char *file, unsigned char hash[16], 
		struct nlpcdbr *rec)
{
	struct nlpcdbr *tmp;
	unsigned char *hashp;

	assert(NULL != file);
	assert(NULL != dbp);
	assert(NULL != rec);

	if ( ! nlpcdbr_nextraw(file, &hashp, &tmp))
		return(0);

	assert(NULL != tmp);
	assert(NULL != hashp);

	nlpcdbr_copyout(rec, tmp);
	memcpy(hash, hashp, 16);

	return(1);
}

static int
nlpcdb_ismatchraw(const struct nlpcdbr *match, 
		const struct nlpcdbr *raw, uint8_t fields)
{
	uint8_t i, j;
	for (i = 0; i < 8; i++) {
		if ( ! (fields & (1 << i)))
			continue;
		switch (1 << i) {
		case (NLPCDBR_FIELD_URI):
			if (0 != memcmp(match->uri, raw->uri, 
						NLPCR_URILEN))
				return(0);
			break;
		case (NLPCDBR_FIELD_STATE):
			for (j = 0; j < 8; j++)
				if ((match->state & (1 << j)) && 
						(raw->state & (1 << j)))
					break;
			if (j == 8)
				return(0);
			break;
		case (NLPCDBR_FIELD_CTIME):
			if (match->ctime != ntohl(raw->ctime))
				return(0);
			break;
		case (NLPCDBR_FIELD_MTIME):
			if (match->mtime != ntohl(raw->mtime))
				return(0);
			break;
		case (NLPCDBR_FIELD_ATIME):
			if (match->atime != ntohl(raw->atime))
				return(0);
			break;
		default:
			abort();
		}
	}
	return(1);
}

/*extern*/ int
nlpcdb_nextmatch(const char *file, 
		unsigned char hash[NLPCR_HASHLEN],
		struct nlpcdbr *rec, const struct nlpcdbr *match,
		uint8_t fields, unsigned long burst)
{
	unsigned long	i;
	struct nlpcdbr 	*tmp;
	unsigned char 	*hashp;

	assert(NULL != file);
	assert(NULL != dbp);
	assert(NULL != rec);

	/* XXX - will this ever max out? */
	burst = (burst == 0 ? ULONG_MAX : burst);

	for (i = 0; i < burst; i++) {
		if ( ! nlpcdbr_nextraw(file, &hashp, &tmp)) 
			return (-1);
		assert(NULL != hashp);
		assert(NULL != tmp);
		if ( ! nlpcdb_ismatchraw(match, tmp, fields)) 
			continue;
		nlpcdbr_copyout(rec, tmp);
		memcpy(hash, hashp, 16);
		break;
	}

	return(i == burst ? 0 : 1);
}

/*extern*/ void
nlpcdb_gethash(const char *uri, 
		unsigned char hash[NLPCR_HASHLEN])
{
	MD5_CTX 	ctx;

	assert(NULL != uri);

	MD5Init(&ctx);
	MD5Update(&ctx, (const unsigned char *)uri, strlen(uri + 1));
	MD5Final(hash, &ctx);
}

/*extern*/ void
nlpcdb_update(const char *file, 
		unsigned char hash[NLPCR_HASHLEN],
		const struct nlpcdbr *rec)
{
	DBC		*dbc;
	DBT		key, data;
	int		rc;

	/*
	 * Note that this may not be atomic, meaning that if another
	 * application changes the database, the cursor may no longer be
	 * valid.  This will go away if modifications use transactions,
	 * which may occur in later versions of this software.
	 */

	assert(NULL != file);
	assert(NULL != rec);
	assert(NULL != dbp);

	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));
	dbc = NULL;

	rc = dbp->cursor(dbp, NULL, &dbc, 0);
	if (0 != rc)
		errx(EXIT_FAILURE, "database cursor: %s, %s",
				file, db_strerror(rc));

	key.data = hash;
	key.size = NLPCR_HASHLEN;

	if (0 != (rc = dbc->c_get(dbc, &key, &data, DB_SET))) 
		errx(EXIT_FAILURE, "database query: %s, %s",
				file, db_strerror(rc));

	nlpcdbr_copyin((struct nlpcdbr *)data.data, rec);

	if (0 != (rc = dbc->c_put(dbc, &key, &data, DB_CURRENT)))
		errx(EXIT_FAILURE, "database put: %s, %s",
				file, db_strerror(rc));
	if (0 != (rc = dbc->c_close(dbc)))
		errx(EXIT_FAILURE, "database close: %s, %s",
				file, db_strerror(rc));
}

/*extern*/ int
nlpcdb_add(const char *file, 
		const unsigned char hash[NLPCR_HASHLEN],
		const struct nlpcdbr *rec)
{
	int 		rc;
	struct nlpcdbr	*p;
	unsigned char	*digest;
	DBT 		key, data;

	assert(NULL != file);
	assert(NULL != rec);
	assert(NULL != dbp);

	if (NULL == (digest = malloc(16)))
		err(EXIT_FAILURE, "failed allocate memory");
	if (NULL == (p = malloc(sizeof(struct nlpcdbr))))
		err(EXIT_FAILURE, "failed allocate memory");

	memcpy(digest, hash, NLPCR_HASHLEN);
	nlpcdbr_copyin(p, rec);

	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	key.data = digest;
	key.size = NLPCR_HASHLEN;

	data.data = p;
	data.size = sizeof(struct nlpcdbr);

	switch ((rc = dbp->put(dbp, NULL, &key, 
				&data, DB_NOOVERWRITE))) {
	case (0):
		return(1);
	case (DB_KEYEXIST):
		free(digest);
		free(p);
		return(0);
	default:
		free(digest);
		free(p);
		err(EXIT_FAILURE, "database modify: %s (%s)",
				file, db_strerror(rc));
	}
	/*NOTREACHED*/
}

/*extern*/ void
nlpcdb_dbg(const char *fmt, ...) 
{
	va_list 	ap;
	extern char 	*__progname;

	if (0 >= nlpcdb_dbg_lvl)
		return;

	va_start(ap, fmt);
	fprintf(stdout, "%s: debug1: ", __progname);
	vprintf(fmt, ap);
	fprintf(stdout, "\n");
	fflush(stdout);
	va_end(ap);
}

/*extern*/ void
nlpcdb_ddbg(const char *fmt, ...) 
{
	va_list 	ap;
	extern char 	*__progname;

	if (1 >= nlpcdb_dbg_lvl)
		return;

	va_start(ap, fmt);
	fprintf(stdout, "%s: debug2: ", __progname);
	vprintf(fmt, ap);
	fprintf(stdout, "\n");
	fflush(stdout);
	va_end(ap);
}

